{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RankNTypes #-}

module Control.Monad.IRTree (
  -- * IRTree
  IRTree,
  extract,
  -- probe,
  reduce,
  reduceExp,

  -- * IRTreeT
  IRTreeT,
  -- probeT,
  extractT,
  reduceT,
  reduceExpT,

  -- * Re-exports
  module Control.Monad.Reader,
  module Data.RPath,
) where

import Control.Monad.RWS.Strict
import Control.Monad.Reader
import Control.Monad.Reduce
import Data.Bits
import Data.Foldable
import Data.Functor
import Data.Functor.Identity
import Data.RPath
import qualified Data.Sequence as Seq

type IRTree l = IRTreeT l Identity

newtype IRTreeT l m i = IRTreeT (RWST RPath (Endo [l]) Int m i)
  deriving
    (Functor, Applicative, Monad)
    via (RWST RPath (Endo [l]) Int m)

instance (Monad m) => MonadReduce l (IRTreeT l m) where
  check l = IRTreeT . RWST $ \rp i -> do
    pure (indexChoice rp i, i + 1, Endo (l :))

extract :: IRTree l i -> i
extract t = runIdentity $ extractT t
{-# INLINE extract #-}

extractT :: (Functor m) => IRTreeT l m i -> m i
extractT (IRTreeT m) = fmap (\(i, _, _) -> i) (runRWST m "" 0)
{-# INLINE extractT #-}

-- probe :: IRTree l i -> RPath -> (i, [(Bool, l)])
-- probe t pth = runIdentity $ probeT t pth
-- {-# INLINE probe #-}

probeT :: (Functor m) => IRTreeT l m i -> RPath -> m (i, [(Bool, l)])
probeT (IRTreeT m) pth =
  runRWST m pth 0 <&> \(i, _, l) ->
    (i, zip (toChoiceList pth) (appEndo l []))
{-# INLINE probeT #-}

reduce
  :: (Monad m)
  => ([(Bool, l)] -> i -> m Bool)
  -> IRTree l i
  -> m (Maybe i)
reduce = reduceT (pure . runIdentity)
{-# INLINE reduce #-}

-- | Interpreted reduction with an m base monad
reduceT
  :: (Monad m, Functor t)
  => (forall a. t a -> m a)
  -> ([(Bool, l)] -> i -> m Bool)
  -> IRTreeT l t i
  -> m (Maybe i)
reduceT lift_ p rt = do
  (i, l) <- lift_ (probeT rt "")
  t <- p l i
  if t
    then Just <$> go Seq.empty
    else pure Nothing
 where
  go pth = do
    -- Try to run the true branch.
    let pth' = fromChoiceList (toList (pth Seq.|> True))
    (i, l) <- lift_ $ probeT rt pth'
    if length l >= numberOfChoices pth'
      then do
        t <- p l i
        go (pth Seq.|> t)
      else pure i
{-# INLINE reduceT #-}

reduceExp
  :: (Monad m)
  => ([(Bool, l)] -> i -> m Bool)
  -> IRTree l i
  -> m (Maybe i)
reduceExp = reduceExpT (pure . runIdentity)
{-# INLINE reduceExp #-}

-- | Interpreted reduction with an m base monad, and running in expoential mode.
reduceExpT
  :: (Monad m, Functor t)
  => (forall a. t a -> m a)
  -- ^ a lift of monad m into t (normally @id@ or @lift@)
  -> ([(Bool, l)] -> i -> m Bool)
  -> IRTreeT l t i
  -> m (Maybe i)
reduceExpT lift_ p rt = do
  (i, l) <- lift_ (probeT rt "")
  t <- p l i
  if t
    then Just <$> go 0 Seq.empty
    else pure Nothing
 where
  go n sq = do
    let depth = shiftL 1 n
    let sq' = sq <> Seq.replicate depth True
    let pth' = fromChoiceList (toList sq')
    (i, l) <- lift_ (probeT rt pth')
    if length l >= numberOfChoices pth' - depth + 1
      then do
        t <- p l i
        if t
          then go (n + 1) sq'
          else case n of
            0 -> go 0 (sq Seq.|> False)
            n' -> go (n' - 1) sq
      else pure i
{-# INLINE reduceExpT #-}
