module Data.RPath (
  -- * RPath
  RPath,
  fromChoiceList,
  toChoiceList,
  numberOfChoices,
  indexChoice,

  -- * As a predicate
  toPredicate,
  toPredicateDebug,

  -- * Helpers
  debugShow,
  debugShowWithDepth,
) where

import Data.Bool
import Data.IORef (atomicModifyIORef, newIORef)
import Data.Maybe
import Data.String
import qualified Data.Vector.Unboxed as VU

{- | A reduction path, can be used as an index into reduction tree.
Is isomorphic to a list of choices.
-}
newtype RPath = RPath {rPathAsVector :: VU.Vector Bool}
  deriving (Eq, Ord)

{- | Convert an RPath into a predicate, which will find the element which
the rpath points to.
-}
toPredicate :: RPath -> IO (i -> IO Bool)
toPredicate rp = do
  idx <- newIORef (-1)
  pure . const $ do
    idx' <- atomicModifyIORef idx (\n -> (n + 1, n))
    pure $ indexChoice rp idx'

-- | Like @toPredicate@, but with debugging information
toPredicateDebug :: (Show i) => RPath -> IO (i -> IO Bool)
toPredicateDebug rp = do
  idx <- newIORef (-1)
  pure $ \i -> do
    idx' <- atomicModifyIORef idx (\n -> (n + 1, n))
    putStr (debugShowWithDepth rp idx')
    putStr ": "
    print i
    pure $ indexChoice rp idx'

{- | Index the list of choices, if there are no choices left
default to false. Indexing with a negative number gives True.
-}
indexChoice :: RPath -> Int -> Bool
indexChoice (RPath v) idx
  | idx < 0 = True
  | otherwise = fromMaybe False (v VU.!? idx)

-- | Get the number of choices in the RPath.
numberOfChoices :: RPath -> Int
numberOfChoices (RPath v) = VU.length v

-- | Create a reduction path from a list of choices
fromChoiceList :: [Bool] -> RPath
fromChoiceList = RPath . VU.fromList

-- | Get a list of choices from a reduction path.
toChoiceList :: RPath -> [Bool]
toChoiceList = VU.toList . rPathAsVector

debugShowWithDepth :: RPath -> Int -> String
debugShowWithDepth rp i =
  (map (bool '0' '1') . take (i + 1) . toChoiceList $ rp)
    <> replicate (numberOfChoices rp - i - 1) '*'

debugShow :: RPath -> String
debugShow = map (bool '0' '1') . toChoiceList

instance Show RPath where
  show = show . debugShow

instance Read RPath where
  readsPrec i s = [(fromString a, r) | (a, r) <- readsPrec i s]

instance IsString RPath where
  fromString = fromChoiceList . map (== '1')
