{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RankNTypes #-}

module Control.Monad.RTreeSpec where

import Control.Monad.Identity (Identity (runIdentity))
import Control.Monad.RTree
import Control.Monad.Reader
import Data.Expr as Expr
import Data.Foldable
import Data.Functor
import qualified Data.Map.Strict as Map
import Test.Hspec
import qualified Test.Hspec.Expectations.Pretty as Pretty
import Test.Hspec.Glitter

shouldBeString :: String -> String -> Pretty.Expectation
shouldBeString = Pretty.shouldBe

rBool :: (MonadReduce () m) => m Bool
rBool = split () (pure False) (pure True)

rList :: (MonadReduce () m) => [Int] -> m [Int]
rList = collect (given () $>)

spec :: Spec
spec = do
  rtreeSpec
  rtreeTSpec
  examplesSpec

rtreeTSpec :: Spec
rtreeTSpec = describe "RTreeT" do
  describe "equivalence" do
    it "should extract like RTree" do
      equiv (rList [1, 2, 3 :: Int]) extract (runIdentity . extractT)

    it "should input like RTree" do
      equiv (rList [1, 2, 3 :: Int]) inputs (toList :: RTreeT () Identity [Int] -> [[Int]])

equiv
  :: (Show b, MonadReduce l x, MonadReduce l y, Eq b)
  => (forall m. (MonadReduce l m) => m a)
  -> (x a -> b)
  -> (y a -> b)
  -> IO ()
equiv tree fn1 fn2 =
  fn1 tree `shouldBe` fn2 tree

rtreeSpec :: Spec
rtreeSpec = describe "RTree" do
  describe "extract" do
    it "should extract rlist" do
      extract (rList [1, 2, 3]) `shouldBe` [1, 2, 3 :: Int]

    it "should extract rBool" do
      extract rBool `shouldBe` True

  describe "inputs" do
    it "should get all inputs from rBool" do
      inputs rBool `shouldBe` [False, True]

    it "should get all inputs from rList" do
      inputs (rList [1, 2, 3 :: Int])
        `shouldBe` [[], [3], [2], [2, 3], [1], [1, 3], [1, 2], [1, 2, 3]]

  describe "iinputs" do
    it "should get all inputs from rList" do
      iinputs (rList [1, 2, 3 :: Int])
        `shouldBe` [ ("111", [])
                   , ("110", [3])
                   , ("101", [2])
                   , ("100", [2, 3])
                   , ("011", [1])
                   , ("010", [1, 3])
                   , ("001", [1, 2])
                   , ("000", [1, 2, 3])
                   ]

  describe "drawRTree" do
    it "should pretty print it's tree" do
      glitter
        "test/expected/rlist-drawrtree"
        (drawRTree (\() -> id) shows (rList [1, 2, 3 :: Int]))

examplesSpec :: Spec
examplesSpec = describe "example" do
  handle "small-opr-expr" $
    Opr (Cnt 1) (Cnt 2)

  handle "small-let-expr" $
    Let "x" (Cnt 1) (Opr (Cnt 2) (Var "x"))

  handle "double-let-expr" $
    Let "x" (Cnt 1) (Let "y" (Cnt 2) (Opr (Var "x") (Var "y")))

  handle "double-overloading-let-expr" $
    Let "x" (Cnt 1) (Let "x" (Cnt 2) (Opr (Var "x") (Var "x")))
 where
  handle str e = describe (str <> " (" <> prettyExprS 0 e ")") do
    let me = Expr.rExpr e
    it "should extract" do
      extract (runReaderT me Map.empty)
        `shouldBe` e

    let re = runReaderT me Map.empty

    it "should draw the same" do
      glitter
        ("test/expected/" <> str)
        (drawRTree showString (prettyExprS 0) re)

    it "should reduce like iinputs" do
      forM_ (iinputs re) \(ii, e') -> do
        p <- toPredicate ii
        reduce p re `shouldReturn` Just e'
